import torch
import torch.nn as nn
import torch.nn.functional as F
import networkx as nx
import numpy as np
import pypaddle.util
from torch.autograd import Variable


class LayeredGraph(nx.DiGraph):
    @property
    def first_layer(self):
        """
        :rtype: int
        """
        return NotImplementedError()

    @property
    def last_layer(self):
        """
        :rtype: int
        """
        return NotImplementedError()

    @property
    def num_layers(self):
        """
        :rtype: int
        """
        return NotImplementedError()

    @property
    def first_layer_size(self):
        """
        :rtype: int
        """
        return NotImplementedError()

    @property
    def last_layer_size(self):
        """
        :rtype: int
        """
        return NotImplementedError()

    @property
    def layers(self):
        """
        :rtype: list[int]
        """
        raise NotImplementedError()

    def get_layer(self, vertex: int):
        """
        :rtype: int
        """
        raise NotImplementedError()

    def get_vertices(self, layer: int):
        """
        :rtype: list[int]
        """
        raise NotImplementedError()

    def get_layer_size(self, layer: int):
        """
        :rtype: int
        """
        raise NotImplementedError()

    def layer_connected(self, layer_index1: int, layer_index2: int):
        """
        :rtype: bool
        """
        raise NotImplementedError()

    def layer_connection_size(self, layer_index1: int, layer_index2: int):
        """
        :rtype: int
        """
        raise NotImplementedError()


class CachedLayeredGraph(LayeredGraph):
    def __init__(self, **attr):
        super(CachedLayeredGraph, self).__init__(**attr)
        self._has_changed = True
        self._layer_index = None
        self._vertex_by_layer = None

    def add_cycle(self, nodes, **attr):
        super(LayeredGraph, self).add_cycle(nodes, **attr)
        self._has_changed = True

    def add_edge(self, u_of_edge, v_of_edge, **attr):
        super(LayeredGraph, self).add_edge(u_of_edge, v_of_edge, **attr)
        self._has_changed = True

    def add_edges_from(self, ebunch_to_add, **attr):
        super(LayeredGraph, self).add_edges_from(ebunch_to_add, **attr)
        self._has_changed = True

    def add_node(self, node_for_adding, **attr):
        super(LayeredGraph, self).add_node(node_for_adding, **attr)
        self._has_changed = True

    def add_nodes_from(self, nodes_for_adding, **attr):
        super(LayeredGraph, self).add_nodes_from(nodes_for_adding, **attr)
        self._has_changed = True

    def add_path(self, nodes, **attr):
        super(LayeredGraph, self).add_path(nodes, **attr)
        self._has_changed = True

    def add_star(self, nodes, **attr):
        super(LayeredGraph, self).add_star(nodes, **attr)
        self._has_changed = True

    def add_weighted_edges_from(self, ebunch_to_add, weight='weight', **attr):
        super(LayeredGraph, self).add_weighted_edges_from(ebunch_to_add, weight='weight', **attr)
        self._has_changed = True

    def _get_layer_index(self):
        if self._has_changed or self._layer_index is None or self._vertex_by_layer is None:
            self._build_layer_index()
            self._has_changed = False

        return self._layer_index, self._vertex_by_layer

    def _layer_by_vertex(self, vertex: int):
        return self._get_layer_index()[0][vertex]

    def _vertices_by_layer(self, layer: int):
        return self._get_layer_index()[1][layer]

    def _build_layer_index(self):
        self._layer_index, self._vertex_by_layer = pypaddle.util.build_layer_index(self)

    @property
    def first_layer(self):
        """
        :rtype: int
        """
        return self.layers[0]

    @property
    def last_layer(self):
        """
        :rtype: int
        """
        return self.layers[-1]

    @property
    def num_layers(self):
        return len(self.layers)

    @property
    def first_layer_size(self):
        return self.get_layer_size(self.layers[0])

    @property
    def last_layer_size(self):
        return self.get_layer_size(self.layers[-1])

    @property
    def layers(self):
        return [layer for layer in self._get_layer_index()[1]]

    def get_layer(self, vertex: int):
        return self._layer_by_vertex(vertex)

    def get_vertices(self, layer: int):
        return self._vertices_by_layer(layer)

    def get_layer_size(self, layer: int):
        return len(self._vertices_by_layer(layer))

    def layer_connected(self, layer_index1: int, layer_index2: int):
        """
        :rtype: bool
        """
        if layer_index1 is layer_index2:
            raise ValueError('Same layer does not have interconnections, it would be split up.')
        if layer_index1 > layer_index2:
            tmp = layer_index2
            layer_index2 = layer_index1
            layer_index1 = tmp

        for source_vertex in self.get_vertices(layer_index1):
            for target_vertex in self.get_vertices(layer_index2):
                if self.has_edge(source_vertex, target_vertex):
                    return True
        return False

    def layer_connection_size(self, layer_index1: int, layer_index2: int):
        """
        :rtype: int
        """
        if layer_index1 is layer_index2:
            raise ValueError('Same layer does not have interconnections, it would be split up.')
        if layer_index1 > layer_index2:
            tmp = layer_index2
            layer_index2 = layer_index1
            layer_index1 = tmp

        size = 0
        for source_vertex in self.get_vertices(layer_index1):
            for target_vertex in self.get_vertices(layer_index2):
                if self.has_edge(source_vertex, target_vertex):
                    size += 1
        return size


class DeepCellDAN(nn.Module):
    def __init__(self, num_classes, input_channel_size, cell_constructor, structure: LayeredGraph):
        super(DeepCellDAN, self).__init__()

        self._structure = structure
        assert structure.num_layers > 0

        self._input_nodes = [n for n in structure.nodes if structure.in_degree(n) == 0]
        self._output_nodes = [n for n in structure.nodes if structure.out_degree(n) == 0]
        assert len(self._input_nodes) > 0, 'No input nodes in structure: len=%d' % len(self.input_nodes)
        assert len(self._output_nodes) > 0, 'No output nodes in structure: len=%d' % len(self.output_nodes)

        self._node_cells = {
            node: cell_constructor(
                is_input=node in self._input_nodes,
                is_output=node in self._output_nodes,
                in_degree=structure.in_degree(node),
                out_degree=structure.out_degree(node),
                layer=structure.get_layer(node)
            )
            for node in self._structure.nodes
        }
        self._nodes = nn.ModuleList(list(self._node_cells.values()))


class MaskableModule(nn.Module):
    def apply_mask(self):
        for layer in maskable_layers(self):
            layer.apply_mask()

    def recompute_mask(self, theta=0.0001):
        for layer in maskable_layers(self):
            layer.recompute_mask(theta)


class MaskedDeepDAN(MaskableModule):
    """
    A deep directed acyclic network model which is capable of masked layers and masked skip-layer connections.
    """
    def __init__(self, input_size, num_classes, structure : LayeredGraph):
        super(MaskedDeepDAN, self).__init__()

        self._structure = structure
        assert structure.num_layers > 0

        # Multiple dimensions for input size are flattened out
        if type(input_size) is tuple or type(input_size) is torch.Size:
            input_size = np.prod(input_size)
        input_size = int(input_size)

        self.layer_first = MaskedLinearLayer(input_size, structure.first_layer_size)
        self.layers_main_hidden = nn.ModuleList([MaskedLinearLayer(structure.get_layer_size(l-1), structure.get_layer_size(l)) for l in structure.layers[1:]])

        for layer_idx, layer in zip(structure.layers[1:], self.layers_main_hidden):
            mask = torch.zeros(structure.get_layer_size(layer_idx), structure.get_layer_size(layer_idx-1))
            for source_idx, source_vertex in enumerate(structure.get_vertices(layer_idx-1)):
                for target_idx, target_vertex in enumerate(structure.get_vertices(layer_idx)):
                    if structure.has_edge(source_vertex, target_vertex):
                        mask[target_idx][source_idx] = 1
            layer.set_mask(mask)

        skip_layers = []
        self._skip_targets = {}
        for target_layer in structure.layers[2:]:
            target_size = structure.get_layer_size(target_layer)
            for distant_source_layer in structure.layers[:target_layer-1]:
                if structure.layer_connected(distant_source_layer, target_layer):
                    if target_layer not in self._skip_targets:
                        self._skip_targets[target_layer] = []

                    skip_layer = MaskedLinearLayer(structure.get_layer_size(distant_source_layer), target_size)
                    mask = torch.zeros(structure.get_layer_size(target_layer), structure.get_layer_size(distant_source_layer))
                    for source_idx, source_vertex in enumerate(structure.get_vertices(distant_source_layer)):
                        for target_idx, target_vertex in enumerate(structure.get_vertices(target_layer)):
                            if structure.has_edge(source_vertex, target_vertex):
                                mask[target_idx][source_idx] = 1
                    skip_layer.set_mask(mask)

                    skip_layers.append(skip_layer)
                    self._skip_targets[target_layer].append({'layer': skip_layer, 'source': distant_source_layer})
        self.layers_skip_hidden = nn.ModuleList(skip_layers)

        self.layer_out = MaskedLinearLayer(structure.last_layer_size, num_classes)
        self.activation = nn.ReLU()

    def generate_structure(self, include_input=False, include_output=False):
        structure = CachedLayeredGraph()

        node_number_offset = 0
        layer_nodeidx2node = {}

        if include_input:
            layer = self.layer_first
            layer_nodeidx2node[0] = {node_idx: node for node_idx, node in enumerate(np.arange(node_number_offset, node_number_offset + layer.mask.shape[1]))}
            node_number_offset += layer.mask.shape[1]
            structure.add_nodes_from(layer_nodeidx2node[0].values())

            for source_node_idx in range(layer.mask.shape[1]):
                source_node = layer_nodeidx2node[0][source_node_idx]

                layer_nodeidx2node[1] = {node_idx: node for node_idx, node in enumerate(np.arange(node_number_offset, node_number_offset + layer.mask.shape[0]))}
                node_number_offset += layer.mask.shape[0]
                structure.add_nodes_from(layer_nodeidx2node[1].values())

                for target_node_idx in range(layer.mask.shape[0]):
                    if layer.mask[target_node_idx][source_node_idx]:
                        target_node = layer_nodeidx2node[1][target_node_idx]
                        structure.add_edge(source_node, target_node)

        # Main hidden layers
        for source_layer_idx, layer in enumerate(self.layers_main_hidden, start=1):
            if source_layer_idx not in layer_nodeidx2node:
                layer_nodeidx2node[source_layer_idx] = {node_idx: node for node_idx, node in enumerate(np.arange(node_number_offset, node_number_offset + layer.mask.shape[1]))}
                node_number_offset += layer.mask.shape[1]
                structure.add_nodes_from(layer_nodeidx2node[source_layer_idx].values())

            target_layer_idx = source_layer_idx+1
            if target_layer_idx not in layer_nodeidx2node:
                layer_nodeidx2node[target_layer_idx] = {node_idx: node for node_idx, node in enumerate(np.arange(node_number_offset, node_number_offset + layer.mask.shape[0]))}
                node_number_offset += layer.mask.shape[0]
                structure.add_nodes_from(layer_nodeidx2node[target_layer_idx].values())

            for source_node_idx in range(layer.mask.shape[1]):
                source_node = layer_nodeidx2node[source_layer_idx][source_node_idx]

                for target_node_idx in range(layer.mask.shape[0]):
                    target_node = layer_nodeidx2node[target_layer_idx][target_node_idx]

                    if layer.mask[target_node_idx][source_node_idx]:
                        structure.add_edge(source_node, target_node)

        # Skip layers
        for target_idx in self._skip_targets:
            for target in self._skip_targets[target_idx]:
                source_layer = target['layer']
                # We have to shift target layer index as layer_nodeidx2node is indexed differently
                target_layer_idx = target_idx + 1
                # Also shift the source layer index by one as we indexed layer_nodeidx2node differently
                source_layer_idx = target['source'] + 1

                if source_layer_idx not in layer_nodeidx2node:
                    # Possibly the source layer (e.g. input layer) was left out, so skip connections from it can not be
                    # considered
                    continue

                source_idx2node = layer_nodeidx2node[source_layer_idx]
                target_idx2node = layer_nodeidx2node[target_layer_idx]

                for source_node_idx in range(source_layer.mask.shape[1]):
                    source_node = source_idx2node[source_node_idx]
                    for target_node_idx in range(source_layer.mask.shape[0]):
                        if source_layer.mask[target_node_idx][source_node_idx]:
                            target_node = target_idx2node[target_node_idx]
                            structure.add_edge(source_node, target_node)

        if include_output:
            layer = self.layer_out
            layer_indices = len(layer_nodeidx2node)
            layer_nodeidx2node[layer_indices] = {node_idx: node for node_idx, node in enumerate(np.arange(node_number_offset, node_number_offset + layer.mask.shape[0]))}
            node_number_offset += layer.mask.shape[0]
            structure.add_nodes_from(layer_nodeidx2node[layer_indices].values())

            for source_node_idx in range(layer.mask.shape[1]):
                source_node = layer_nodeidx2node[layer_indices-1][source_node_idx]
                for target_node_idx in range(layer.mask.shape[0]):
                    if layer.mask[target_node_idx][source_node_idx]:
                        target_node = layer_nodeidx2node[layer_indices][target_node_idx]
                        structure.add_edge(source_node, target_node)

        return structure

    def forward(self, input):
        # input : [batch_size, ?, ?, ..], e.g. [100, 1, 28, 28] or [100, 3, 32, 32]
        x = input.flatten(start_dim=1)
        # x: [batch_size, ?*?*..], e.g. [100, 784] or [100, 3072]
        last_output = self.activation(self.layer_first(x))
        layer_results = {0: last_output}
        for layer, layer_idx in zip(self.layers_main_hidden, self._structure.layers[1:]):
            out = self.activation(layer(last_output))

            if layer_idx in self._skip_targets:
                for skip_target in self._skip_targets[layer_idx]:
                    source_layer = skip_target['layer']
                    source_idx = skip_target['source']

                    out += self.activation(source_layer(layer_results[source_idx]))

            layer_results[layer_idx] = out  # copy?
            last_output = out

        return self.layer_out(last_output)


class MaskedDeepFFN(MaskableModule):
    """
    A deep feed-forward network model which is capable of masked layers.
    Masked layers can represent sparse structures between consecutive layers.
    This representation is suitable for feed-forward sparse networks, probably with density 0.5 and above per layer.
    """
    def __init__(self, input_size, num_classes, hidden_layers : list):
        super(MaskedDeepFFN, self).__init__()
        assert len(hidden_layers) > 0

        # Multiple dimensions for input size are flattened out
        if type(input_size) is tuple or type(input_size) is torch.Size:
            input_size = np.prod(input_size)
        input_size = int(input_size)

        self.layer_first = MaskedLinearLayer(input_size, hidden_layers[0])
        self.layers_hidden = nn.ModuleList([MaskedLinearLayer(hidden_layers[l], h) for l, h in enumerate(hidden_layers[1:])])
        self.layer_out = MaskedLinearLayer(hidden_layers[-1], num_classes)
        self.activation = nn.ReLU()

    def generate_structure(self, include_input=False, include_output=False):
        """

        :param include_input:
        :param include_output:
        :rtype : LayeredGraph
        :return:
        """
        structure = CachedLayeredGraph()

        def add_edges(structure, layer, offset_source):
            offset_target = offset_source + layer.mask.shape[1]
            for source_node_idx, source_node in enumerate(range(layer.mask.shape[1])):
                for target_node_idx, target_node in enumerate(range(layer.mask.shape[0])):
                    if layer.mask[target_node_idx][source_node_idx]:
                        structure.add_edge(offset_source + source_node, offset_target + target_node)
            return offset_target

        offset_source = 0
        if include_input:
            offset_source = add_edges(structure, self.layer_first, 0)

        for layer in self.layers_hidden:
            offset_source = add_edges(structure, layer, offset_source)

        if include_output:
            add_edges(structure, self.layer_out, offset_source)

        return structure

    def forward(self, input):
        # input : [batch_size, ?, ?, ..], e.g. [100, 1, 28, 28] or [100, 3, 32, 32]
        out = self.activation(self.layer_first(input.flatten(start_dim=1)))  # [B, n_hidden_1]
        for layer in self.layers_hidden:
            out = self.activation(layer(out))
        return self.layer_out(out)  # [B, n_out]

    def to(self, *args, **kwargs):
        super().to(*args, **kwargs)
        self.layer_first.to(*args, **kwargs)
        for idx, h in enumerate(self.layers_hidden):
            self.layers_hidden[idx] = self.layers_hidden[idx].to(*args, **kwargs)
        self.layer_out.to(*args, **kwargs)
        return self


def maskable_layers(network):
    for child in network.children():
        if type(child) is MaskedLinearLayer:
            yield child
        elif type(child) is nn.ModuleList:
            for layer in maskable_layers(child):
                yield layer


def maskable_layers_with_name(network):
    for name, child in network.named_children():
        if type(child) is MaskedLinearLayer:
            yield name, child
        elif type(child) is nn.ModuleList:
            for name, layer in maskable_layers_with_name(child):
                yield name, layer


def prunable_layers(network):
    return maskable_layers(network)


def prunable_layers_with_name(network):
    return maskable_layers_with_name(network)


class MaskedLinearLayer(nn.Linear):
    def __init__(self, in_feature, out_features, bias=True, keep_layer_input=False):
        """
        :param in_feature:          The number of features that are inserted in the layer.
        :param out_features:        The number of features that are returned by the layer.
        :param bias:                Iff each neuron in the layer should have a bias unit as well.
        """
        super().__init__(in_feature, out_features, bias)

        self.register_buffer('mask', torch.ones((out_features, in_feature), dtype=torch.bool))
        self.keep_layer_input = keep_layer_input
        self.layer_input = None

    def get_mask(self):
        return self.mask

    def set_mask(self, mask):
        self.mask = Variable(mask)

    def get_weight_count(self):
        return self.mask.sum()

    def apply_mask(self):
        # Assigning "self.weight = torch.nn.Parameter(self.weight.mul(self.mask))" might have side effects
        # Using direct manipulation on tensor "self.weight.data"
        self.weight.data = self.weight * self.mask

    def recompute_mask(self, theta: float = 0.001):
        """
        Recomputes the mask based on the weight magnitudes.
        If you want to consider the existing mask, make sure to first call apply_mask() and then recompute it.

        :param theta: Specifies a possible threshold for absolute distance to zero.
        """
        self.mask = torch.ones(self.weight.shape, dtype=torch.bool, device=self.mask.device)
        self.mask[torch.where(abs(self.weight) < theta)] = False

    def get_weight(self):
        return self.weight

    def reset_parameters(self, keep_mask=False):
        super().reset_parameters()
        # hasattr() is necessary because reset_parameters() is called in __init__ of Linear(), but buffer 'mask'
        # may only be registered after super() call, thus 'mask' might not be defined as buffer / attribute, yet
        if hasattr(self, 'mask') and not keep_mask:
            self.mask = torch.ones(self.weight.size(), dtype=torch.bool)

    def forward(self, input):
        x = input.float()  # In case we get a double-tensor passed, force it to be float for multiplications to work

        # Possibly store the layer input
        if self.keep_layer_input:
            self.layer_input = x.data

        return F.linear(x, self.weight * self.mask, self.bias)





import itertools

import torch
import torch.utils
import unittest
import pypaddle.sparse
import numpy as np
import shutil

from pypaddle.util import get_mnist_loaders
from pypaddle.learning import train, test


class FFN2DANTest(unittest.TestCase):
    def setUp(self):
        self.possible_dataset_roots = ['/media/data/set/mnist', 'data/set/mnist']
        self.batch_size = 100
        self.train_loader, self.test_loader, _, self.dataset_root = get_mnist_loaders(self.batch_size, self.possible_dataset_roots)
        self.device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        feature, labels = iter(self.train_loader).next()
        self.input_shape = feature.shape[1:]
        self.output_size = int(labels.shape[-1])

    def tearDown(self):
        if self.dataset_root is not self.possible_dataset_roots[0]:
            print('Deleting', self.dataset_root)
            shutil.rmtree(self.dataset_root)

    def test_transfer_random_reconnected_structure(self):
        loss = torch.nn.CrossEntropyLoss()

        num_epochs = 2
        model = pypaddle.sparse.MaskedDeepFFN(self.input_shape, self.output_size, [100, 50, 20])
        model.to(self.device)
        optimizer = torch.optim.SGD(model.parameters(), lr=0.01)

        for epoch in range(num_epochs):
            train(self.train_loader, model, optimizer, loss, self.device)

        self.assertGreater(test(self.test_loader, model, self.device), 1/self.output_size)

        model.apply_mask()
        model.recompute_mask(theta=0.01)
        structure = model.generate_structure(include_input=False, include_output=False)

        for source_layer in structure.layers:
            source_layer_size = structure.get_layer_size(source_layer)
            for target_layer in structure.layers[source_layer+1:]:
                target_layer_size = structure.get_layer_size(target_layer)

                random_source_nodes = np.random.choice(
                    structure.get_vertices(source_layer),
                    np.random.randint(1, source_layer_size+1),
                    replace=False
                )
                random_target_nodes = np.random.choice(
                    structure.get_vertices(target_layer),
                    np.random.randint(1, target_layer_size+1),
                    replace=False
                )

                structure.add_edges_from([edge for edge in itertools.product(random_source_nodes, random_target_nodes)])

        dan_model = pypaddle.sparse.MaskedDeepDAN(self.input_shape, self.output_size, structure)
        dan_model.to(self.device)
        dan_optimizer = torch.optim.SGD(dan_model.parameters(), lr=0.01)

        for epoch in range(num_epochs):
            train(self.train_loader, dan_model, dan_optimizer, loss, self.device)

        self.assertGreater(test(self.test_loader, dan_model, self.device), 1/self.output_size)



import torch
import torch.utils
import unittest
import pypaddle.sparse
import numpy as np


class MaskedLinearLayerTest(unittest.TestCase):
    def test_get_structure(self):
        model = pypaddle.sparse.MaskedDeepFFN(784, 10, [20, 15, 12])
        structure = model.generate_structure(include_input=True, include_output=True)
        print(structure)
        # TODO

    def test_generate_large_structure(self):
        layers = [1000, 500, 500, 200, 100]
        model = pypaddle.sparse.MaskedDeepFFN((1, 28, 28), 10, layers)
        structure = model.generate_structure()

        self.assertEqual(len(layers), structure.num_layers)
        structure_layer_sizes = [structure.get_layer_size(l) for l in structure.layers]
        for l1, l2 in zip(structure_layer_sizes, layers):
            self.assertEqual(l1, l2, "Structure %s did not match definition %s" % (structure_layer_sizes, layers))

    def test_random_forward_possibly_on_gpu_success(self):
        device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

        # Arrange
        batch_size = 10
        input_size = 784
        output_size = 10
        model = pypaddle.sparse.MaskedDeepFFN(input_size, output_size, [200, 100, 50])
        model.to(device)
        random_input = torch.tensor(np.random.random((batch_size, input_size)), device=device, requires_grad=False)

        # Act
        output = model(random_input)

        # Assert
        self.assertEqual(output.numel(), batch_size*output_size)

    def test_random_forward_with_multiple_dimensions_success(self):
        device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

        # Arrange
        batch_size = 10
        input_size = (10, 5, 8)
        output_size = 10
        model = pypaddle.sparse.MaskedDeepFFN(input_size, output_size, [100, 200, 50])
        model.to(device)
        random_input = torch.tensor(np.random.random((batch_size,)+input_size), device=device, requires_grad=False)

        # Act
        output = model(random_input)

        # Assert
        self.assertEqual(output.numel(), batch_size*output_size)


import unittest
import pypaddle.util
import networkx as nx


class LayerIndexTest(unittest.TestCase):

    def test_dev(self):
        nodes = [1, 2, 3, 4, 5]

        structure = nx.DiGraph()
        structure.add_nodes_from(nodes)
        structure.add_edge(1, 3)
        structure.add_edge(1, 4)
        structure.add_edge(1, 5)
        structure.add_edge(2, 3)
        structure.add_edge(2, 4)
        structure.add_edge(3, 5)
        structure.add_edge(4, 5)

        layer_index, vertex_by_layer = pypaddle.util.build_layer_index(structure)

        for n in nodes:
            self.assertTrue(n in layer_index)
            self.assertTrue(layer_index[n] in vertex_by_layer)
            self.assertTrue(n in vertex_by_layer[layer_index[n]])



import math
import torch
import torch.nn as nn
import unittest

import pypaddle.deprecated
import pypaddle.sparse
import pypaddle.util
from pypaddle.learning import train, test

class DeepCellDANTest(unittest.TestCase):
    def test_development(self):
        return

        train_loader, test_loader, val_loader = pypaddle.util.get_cifar10_loaders()

        class ReductionCell(nn.Module):
            def __init__(self, input_channels, output_channels):
                super().__init__()

                self.conv_reduce = nn.Conv2d(input_channels, output_channels, kernel_size=5, padding=2, stride=2)
                self.act = nn.ReLU()
                self.batch_norm = nn.BatchNorm2d(output_channels)

            def forward(self, input):
                return self.batch_norm(self.act(self.conv_reduce(input)))

        class Foo(nn.Module):
            def __init__(self, num_classes: int, input_channels: int, input_size: int):
                super(Foo, self).__init__()

                self._num_classes = num_classes

                reduced_channels = 100
                reduction_steps = max(math.floor(math.log2(input_size))-5, 1)
                input_convs = [ReductionCell(input_channels, reduced_channels)]
                for step in range(1, reduction_steps):
                    input_convs.append(ReductionCell(reduced_channels, reduced_channels))
                self._input_convs = nn.ModuleList(input_convs)

                self.conv1 = nn.Conv2d(reduced_channels, 100, (1, 1))
                self.bn1 = nn.BatchNorm2d(100)
                self.fc1 = nn.Linear(100, self._num_classes)

            def forward(self, input):
                y = input
                for conv in self._input_convs:
                    y = conv(y)

                y = self.conv1(y)
                y = self.bn1(y) # [B, X, N, M]
                y = torch.nn.functional.adaptive_avg_pool2d(y, (1, 1)) # [B, X, 1, 1]
                y = y.view(y.size(0), -1) # [B, X]
                return self.fc1(y)  # [B, _num_classes]

        model = Foo(10, 3, 32)
        print(model)
        print([p.shape for p in model.parameters()])

        def count_parameters(model):
            return sum(p.numel() for p in model.parameters() if p.requires_grad)

        device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        print(device)
        model.to(device)

        optimizer = torch.optim.Adam(model.parameters(), lr=0.01, betas=(0.9, 0.999), eps=1e-8, weight_decay=0)
        loss_func = nn.CrossEntropyLoss()

        for epoch in range(50):
            print('Epoch', epoch)
            model.train()
            print(train(train_loader, model, optimizer, loss_func, device))
            model.eval()
            print(test(test_loader, model, device))

        print('Done')
        print('Test:')
        model.eval()
        print(test(test_loader, model, device))

        print('Validation:')
        model.eval()
        print(test(val_loader, model, device))


import os
import math
import torch
import torch.nn as nn
import networkx as nx
import numpy as np
import torchvision

import pypaddle.pruning
import pypaddle.sparse

from torch.utils.data import SubsetRandomSampler


def get_cifar10_loaders(batch_size:int = 100, possible_dataset_roots = 'data/set/cifar10'):
    transform = torchvision.transforms.Compose([
        torchvision.transforms.ToTensor(),
        torchvision.transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))
    ])

    if type(possible_dataset_roots) is str:
        possible_dataset_roots = [possible_dataset_roots]
    assert type(possible_dataset_roots) is list
    assert len(possible_dataset_roots) > 0

    selected_root = possible_dataset_roots[0]
    for possible_root in possible_dataset_roots:
        if os.path.exists(possible_root):
            selected_root = possible_root
    if not os.path.exists(selected_root):
        os.makedirs(selected_root)

    train_set = torchvision.datasets.CIFAR10(root=selected_root, train=True, download=True, transform=transform)
    test_set = torchvision.datasets.CIFAR10(root=selected_root, train=False, download=True, transform=transform)
    n_training_samples = 20000
    train_sampler = SubsetRandomSampler(np.arange(n_training_samples, dtype=np.int64))
    n_val_samples = 5000
    val_sampler = SubsetRandomSampler(np.arange(n_training_samples, n_training_samples + n_val_samples, dtype=np.int64))
    n_test_samples = 5000
    test_sampler = SubsetRandomSampler(np.arange(n_test_samples, dtype=np.int64))

    train_loader = torch.utils.data.DataLoader(train_set, batch_size=batch_size, sampler=train_sampler, num_workers=2)
    test_loader = torch.utils.data.DataLoader(test_set, batch_size=batch_size, sampler=test_sampler, num_workers=2)
    val_loader = torch.utils.data.DataLoader(train_set, batch_size=batch_size, sampler=val_sampler, num_workers=2)

    return train_loader, test_loader, val_loader, selected_root


def get_mnist_loaders(batch_size:int = 100, possible_dataset_roots = 'data/set/mnist'):
    custom_transform = torchvision.transforms.Compose([
        torchvision.transforms.ToTensor(),  # first, convert image to PyTorch tensor
        torchvision.transforms.Normalize((0.1307,), (0.3081,))  # normalize inputs
    ])

    if type(possible_dataset_roots) is str:
        possible_dataset_roots = [possible_dataset_roots]
    assert type(possible_dataset_roots) is list
    assert len(possible_dataset_roots) > 0

    selected_root = possible_dataset_roots[0]
    for possible_root in possible_dataset_roots:
        if os.path.exists(possible_root):
            selected_root = possible_root
    if not os.path.exists(selected_root):
        os.makedirs(selected_root)

    train_set = torchvision.datasets.MNIST(root=selected_root, download=True, train=True, transform=custom_transform)
    test_set = torchvision.datasets.MNIST(root=selected_root, download=True, train=False, transform=custom_transform)

    mnist_train_loader = torch.utils.data.DataLoader(train_set, batch_size=batch_size, shuffle=True, num_workers=2)
    mnist_test_loader = torch.utils.data.DataLoader(test_set, batch_size=batch_size, shuffle=True, num_workers=2)

    return mnist_train_loader, mnist_test_loader, None, selected_root


def build_layer_index(graph : nx.DiGraph, layer_index=None):
    """

    :param graph:
    :type graph igraph.Graph
    :param layer_index:
    :return:
    """
    if layer_index is None:
        layer_index = {}

    recursion_call = {'count': 0}

    def get_layer_index(vertex, graph):
        try:
            vertex = int(vertex)
        except TypeError:
            raise ValueError('You have to pass vertex indices to this function.')
        # print('get_layer_index(%s, roots, graph)' % vertex)
        if vertex is None:
            raise ValueError('Given vertex was none.')
        if vertex not in layer_index:
            recursion_call['count'] += 1
            layer_index[vertex] = max(
                [get_layer_index(v, graph) for v in nx.algorithms.dag.ancestors(graph, vertex)] + [-1]) + 1
        return layer_index[vertex]

    for v in graph:
        get_layer_index(v, graph)

    # print('Recursion call: %s' % recursion_call['count'])

    vertices_by_layer = {}
    for v in layer_index:
        idx = layer_index[v]
        if idx not in vertices_by_layer:
            vertices_by_layer[idx] = []
        vertices_by_layer[idx].append(v)

    return layer_index, vertices_by_layer



"""
This file contains some utility functions to calculate hessian matrix and its inverse.

Adapted from https://github.com/csyhhu/L-OBS/blob/master/PyTorch/ImageNet/util.py
Author: Chen Shangyu (schen025@e.ntu.edu.sg)
"""


def generate_hessian_inverse_fc(layer, hessian_inverse_path, layer_input_train_dir):
    """
    This function calculate hessian inverse for a fully-connect layer
    :param hessian_inverse_path: the hessian inverse path you store
    :param layer: the layer weights
    :param layer_input_train_dir: layer inputs in the dir
    :return:
    """

    w_layer = layer.get_weight().data.numpy().T
    n_hidden_1 = w_layer.shape[0]

    # Here we use a recursive way to calculate hessian inverse
    hessian_inverse = 1000000 * np.eye(n_hidden_1)

    dataset_size = 0
    for input_index, input_file in enumerate(os.listdir(layer_input_train_dir)):
        layer2_input_train = np.load(layer_input_train_dir + '/' + input_file)

        if input_index == 0:
            dataset_size = layer2_input_train.shape[0] * len(os.listdir(layer_input_train_dir))

        for i in range(layer2_input_train.shape[0]):
            # vect_w_b = np.vstack((np.array([layer2_input_train[i]]).T, np.array([[1.0]])))
            vect_w = np.array([layer2_input_train[i]]).T
            denominator = dataset_size + np.dot(np.dot(vect_w.T, hessian_inverse), vect_w)
            numerator = np.dot(np.dot(hessian_inverse, vect_w), np.dot(vect_w.T, hessian_inverse))
            hessian_inverse = hessian_inverse - numerator * (1.00 / denominator)

    np.save(hessian_inverse_path, hessian_inverse)


def edge_cut(layer, hessian_inverse_path, value, strategy=pypaddle.pruning.PruningStrategy.PERCENTAGE):
    """
    This function prune weights of biases based on given hessian inverse and cut ratio
    :param hessian_inverse_path:
    :param layer:
    :param value: The zeros percentage of weights and biases, or, 1 - compression ratio
    :param strategy:
    :return:
    """
    # dataset_size = layer2_input_train.shape[0]
    w_layer = layer.get_weight().data.numpy().T

    # biases = layer.bias.data.numpy()
    n_hidden_1 = w_layer.shape[0]
    n_hidden_2 = w_layer.shape[1]

    sensitivity = np.array([])

    hessian_inverse = np.load(hessian_inverse_path)

    gate_w = layer.get_mask().data.numpy().T
    # gate_b = np.ones([n_hidden_2])

    # calculate number of pruneable elements
    if strategy is pypaddle.pruning.PruningStrategy.PERCENTAGE:
        cut_ratio = value / 100  # transfer percentage from full value to floating point
        max_pruned_num = math.floor(layer.get_weight_count() * cut_ratio)
    elif strategy is pypaddle.pruning.PruningStrategy.BUCKET:
        max_pruned_num = value
    else:
        raise ValueError('Currently not implemented')

    # Calculate sensitivity score. Refer to Eq.5.
    for i in range(n_hidden_2):
        sensitivity = np.hstack(
            (sensitivity, 0.5 * ((w_layer.T[i] ** 2) / np.diag(hessian_inverse))))
    sorted_index = np.argsort(sensitivity)
    hessian_inverseT = hessian_inverse.T

    prune_count = 0
    for i in range(n_hidden_1 * n_hidden_2):
        prune_index = [sorted_index[i]]
        x_index = math.floor(prune_index[0] / n_hidden_1)  # next layer num
        y_index = prune_index[0] % n_hidden_1  # this layer num

        if gate_w[y_index][x_index] == 1:
            delta_w = (-w_layer[y_index][x_index] / hessian_inverse[y_index][y_index]) * hessian_inverseT[y_index]
            gate_w[y_index][x_index] = 0

            if strategy is strategy.PERCENTAGE:
                prune_count += 1
            elif strategy is strategy.BUCKET:
                prune_count += sensitivity[prune_index]  # todo: evaluate here probably a bit wrong :(

            # Parameters update, refer to Eq.5
            w_layer.T[x_index] = w_layer.T[x_index] + delta_w
            # b_layer[x_index] = b_layer[x_index] + delta_w[-1]

        w_layer = w_layer * gate_w
        # b_layer = b_layer * gate_b

        if prune_count == max_pruned_num:
            break

    # set created mask to network again and update the weights
    layer.set_mask(torch.from_numpy(gate_w.T))
    layer.weight = torch.nn.Parameter(torch.from_numpy(w_layer.T))

    # if not os.path.exists(prune_save_path):
    #    os.makedirs(prune_save_path)

    # np.save("%s/weights" % prune_save_path, w_layer)
    # np.save("%s/biases" % prune_save_path, b_layer)


# -----------

def find_network_threshold(network, value, strategy):
    all_sal = []
    for layer in pypaddle.sparse.prunable_layers(network):
        # flatten both weights and mask
        mask = list(layer.get_mask().abs().numpy().flatten())
        saliency = list(layer.get_saliency().numpy().flatten())

        # zip, filter, unzip the two lists
        _, filtered_saliency = zip(
            *((masked_val, weight_val) for masked_val, weight_val in zip(mask, saliency) if masked_val == 1))
        # add all saliencies to list
        all_sal += filtered_saliency

    # calculate percentile
    if strategy is pypaddle.pruning.PruningStrategy.PERCENTAGE:
        return np.percentile(np.array(all_sal), value)
    elif strategy is pypaddle.pruning.PruningStrategy.ABSOLUTE:
        # check if there are enough elements to prune
        if value >= len(all_sal):
            return np.argmax(np.array(all_sal)).item() + 1
        else:
            # determine threshold
            index = np.argsort(np.array(all_sal))[value]
            return np.array(all_sal)[index].item()
    elif strategy is pypaddle.pruning.PruningStrategy.BUCKET:
        sorted_array = np.sort(all_sal)
        sum_array = 0

        # sum up the elements from the sorted array
        for i in sorted_array:
            sum_array += i
            if sum_array > value:
                # return the last element that has been added to the array
                return i

        # return last element if bucket not reached
        return sorted_array[-1] + 1



def set_random_saliency(network):
    # set saliency to random values
    for layer in pypaddle.sparse.prunable_layers(network):
        layer.set_saliency(torch.rand_like(layer.get_weight()) * layer.get_mask())

def set_random_masks(module : nn.Module):
    if isinstance(module, pypaddle.sparse.MaskedLinearLayer):
        module.set_mask(torch.round(torch.rand_like(module.get_weight())))

def set_distributed_saliency(network):
    # prune from each layer the according number of elements
    for layer in pypaddle.sparse.prunable_layers(network):
        # calculate standard deviation for the layer
        w = layer.get_weight().data
        st_v = 1 / w.std()
        # set the saliency in the layer = weight/st.deviation
        layer.set_saliency(st_v * w.abs())


def get_filtered_saliency(saliency, mask):
    s = list(saliency)
    m = list(mask)

    _, filtered_w = zip(
        *((masked_val, weight_val) for masked_val, weight_val in zip(m, s) if masked_val == 1))
    return filtered_w


def get_layer_count(network):
    i = 0
    for _ in pypaddle.sparse.prunable_layers_with_name(network):
        i += 1
    return i


def get_weight_distribution(network):
    all_weights = []
    for layer in pypaddle.sparse.prunable_layers(network):
        mask = list(layer.get_mask().numpy().flatten())
        weights = list(layer.get_weight().data.numpy().flatten())

        masked_val, filtered_weights = zip(
            *((masked_val, weight_val) for masked_val, weight_val in zip(mask, weights) if masked_val == 1))

        all_weights += list(filtered_weights)

    # return all the weights, that are not masked as a numpy array
    return np.array(all_weights)


def get_network_weight_count(network):
    total_weights = 0
    for layer in pypaddle.sparse.prunable_layers(network):
        total_weights += layer.get_weight_count()
    return total_weights


def reset_pruned_network(network):
    for layer in pypaddle.sparse.prunable_layers(network):
        layer.reset_parameters(keep_mask=True)


def keep_input_layerwise(network):
    for layer in pypaddle.sparse.prunable_layers(network):
        layer.keep_layer_input = True



import unittest
import numpy as np
import pypaddle.sparse


class CachedLayeredGraphTest(unittest.TestCase):

    def test_default(self):
        layered_graph = pypaddle.sparse.CachedLayeredGraph()

        layered_graph.add_nodes_from(np.arange(1, 7))

        # First layer
        layered_graph.add_edge(1, 3)
        layered_graph.add_edge(1, 4)
        layered_graph.add_edge(1, 5)
        layered_graph.add_edge(1, 6)
        layered_graph.add_edge(2, 3)
        layered_graph.add_edge(2, 4)
        layered_graph.add_edge(2, 5)
        layered_graph.add_edge(2, 7)

        # Second layer
        layered_graph.add_edge(3, 6)
        layered_graph.add_edge(4, 6)
        layered_graph.add_edge(4, 7)
        layered_graph.add_edge(5, 7)

        first_layer_size_before = layered_graph.get_layer_size(0)
        self.assertFalse(layered_graph._has_changed)

        # Add vertex 0 and connect it to vertices from layer 2
        layered_graph.add_edge(0, 3)
        layered_graph.add_edge(0, 4)
        self.assertTrue(layered_graph._has_changed)

        first_layer_size_after = layered_graph.get_layer_size(0)
        self.assertNotEqual(first_layer_size_before, first_layer_size_after)



import unittest
import numpy as np
import pypaddle.util
import pypaddle.sparse
import networkx as nx

class MaskedDeepDANTest(unittest.TestCase):
    def test_traversal(self):
        # Arrange
        random_graph = nx.watts_strogatz_graph(200, 3, 0.8)
        structure = pypaddle.sparse.CachedLayeredGraph()
        structure.add_edges_from(random_graph.edges)
        structure.add_nodes_from(random_graph.nodes)

        print([structure.in_degree(n) for n in structure.nodes])

    def test_random_structures_success(self):
        # Arrange
        random_graph = nx.watts_strogatz_graph(200, 3, 0.8)
        structure = pypaddle.sparse.CachedLayeredGraph()
        structure.add_edges_from(random_graph.edges)
        structure.add_nodes_from(random_graph.nodes)
        model = pypaddle.sparse.MaskedDeepDAN(784, 10, structure)

        # Act
        extracted_structure = model.generate_structure()
        new_model = pypaddle.sparse.MaskedDeepDAN(784, 10, extracted_structure)

        # Assert
        #self.assertTrue(nx.algorithms.isomorphism.faster_could_be_isomorphic(structure, extracted_structure))
        self.assertTrue(nx.is_isomorphic(structure, extracted_structure))
        self.assertTrue(nx.is_isomorphic(structure, new_model.generate_structure()))

    def test_random_structures_with_input_and_output_success(self):
        # Arrange
        random_graph = nx.watts_strogatz_graph(200, 3, 0.8)
        structure = pypaddle.sparse.CachedLayeredGraph()
        structure.add_edges_from(random_graph.edges)
        structure.add_nodes_from(random_graph.nodes)
        model = pypaddle.sparse.MaskedDeepDAN(784, 10, structure)

        # Act
        extracted_structure = model.generate_structure(include_input=True, include_output=True)
        new_model = pypaddle.sparse.MaskedDeepDAN(784, 10, extracted_structure)

    def test_apply_mask_success(self):
        random_graph = nx.watts_strogatz_graph(200, 3, 0.8)
        structure = pypaddle.sparse.CachedLayeredGraph()
        structure.add_edges_from(random_graph.edges)
        structure.add_nodes_from(random_graph.nodes)
        model = pypaddle.sparse.MaskedDeepDAN(784, 10, structure)

        previous_weights = []
        for layer in pypaddle.sparse.maskable_layers(model):
            previous_weights.append(np.copy(layer.weight.detach().numpy()))

        model.apply_mask()

        different = []
        for layer, previous_weight in zip(pypaddle.sparse.maskable_layers(model), previous_weights):
            different.append(not np.all(np.equal(np.array(previous_weight), np.array(layer.weight.detach().numpy()))))
        self.assertTrue(np.any(different))


    def test_get_structure(self):
        structure = pypaddle.sparse.CachedLayeredGraph()

        block0_size = 8
        block1_size = 8
        block2_size = 2
        block3_size = 2
        block4_size = 2
        block5_size = 2
        block6_size = 10
        block0 = np.arange(1, block0_size+1)
        block1 = np.arange(block0_size+1, block0_size+block1_size+1)
        block2 = np.arange(block0_size+block1_size+1, block0_size+block1_size+block2_size+1)
        block3 = np.arange(block0_size+block1_size+block2_size+1, block0_size+block1_size+block2_size+block3_size+1)
        block4 = np.arange(block0_size+block1_size+block2_size+block3_size+1, block0_size+block1_size+block2_size+block3_size+block4_size+1)
        block5 = np.arange(block0_size+block1_size+block2_size+block3_size+block4_size+1, block0_size+block1_size+block2_size+block3_size+block4_size+block5_size+1)
        block6 = np.arange(block0_size+block1_size+block2_size+block3_size+block4_size+block5_size+1, block0_size+block1_size+block2_size+block3_size+block4_size+block5_size+block6_size+1)

        # First layer
        for v in block0:
            for t in block2:
                structure.add_edge(v, t)
        for v in block0:
            for t in block3:
                structure.add_edge(v, t)
        for v in block0:
            for t in block5:
                structure.add_edge(v, t)
        for v in block1:
            for t in block3:
                structure.add_edge(v, t)
        for v in block1:
            for t in block4:
                structure.add_edge(v, t)
        for v in block1:
            for t in block6:
                structure.add_edge(v, t)

        # Second layer
        for v in block2:
            for t in block5:
                structure.add_edge(v, t)
        for v in block3:
            for t in block5:
                structure.add_edge(v, t)
        for v in block3:
            for t in block6:
                structure.add_edge(v, t)
        for v in block4:
            for t in block6:
                structure.add_edge(v, t)

        model = pypaddle.sparse.MaskedDeepDAN(784, 10, structure)
        print(model)

        new_structure = model.generate_structure(include_input=False, include_output=False)

        model2 = pypaddle.sparse.MaskedDeepDAN(784, 10, new_structure)
        print(model2)

    def test_dev(self):
        structure = pypaddle.sparse.CachedLayeredGraph()
        structure.add_nodes_from(np.arange(1, 7))

        block0_size = 50
        block1_size = 50
        block2_size = 30
        block3_size = 30
        block4_size = 30
        block5_size = 20
        block6_size = 20
        block0 = np.arange(1, block0_size+1)
        block1 = np.arange(block0_size+1, block0_size+block1_size+1)
        block2 = np.arange(block0_size+block1_size+1, block0_size+block1_size+block2_size+1)
        block3 = np.arange(block0_size+block1_size+block2_size+1, block0_size+block1_size+block2_size+block3_size+1)
        block4 = np.arange(block0_size+block1_size+block2_size+block3_size+1, block0_size+block1_size+block2_size+block3_size+block4_size+1)
        block5 = np.arange(block0_size+block1_size+block2_size+block3_size+block4_size+1, block0_size+block1_size+block2_size+block3_size+block4_size+block5_size+1)
        block6 = np.arange(block0_size+block1_size+block2_size+block3_size+block4_size+block5_size+1, block0_size+block1_size+block2_size+block3_size+block4_size+block5_size+block6_size+1)

        # First layer
        for v in block0:
            for t in block2:
                structure.add_edge(v, t)
        for v in block0:
            for t in block3:
                structure.add_edge(v, t)
        for v in block1:
            for t in block3:
                structure.add_edge(v, t)
        for v in block1:
            for t in block4:
                structure.add_edge(v, t)

        # Second layer
        for v in block2:
            for t in block5:
                structure.add_edge(v, t)
        for v in block3:
            for t in block5:
                structure.add_edge(v, t)
        for v in block3:
            for t in block6:
                structure.add_edge(v, t)
        for v in block4:
            for t in block6:
                structure.add_edge(v, t)

        model = pypaddle.sparse.MaskedDeepDAN(784, 10, structure)


import torch
import torch.nn as nn
import unittest
import shutil

import pypaddle.deprecated
import pypaddle.sparse
import pypaddle.util
import networkx as nx
from pypaddle.learning import train, test

class DeepDACellNetworkTest(unittest.TestCase):
    def setUp(self):
        self.possible_dataset_roots = ['/media/data/set/mnist', 'data/set/mnist']
        self.batch_size = 100
        self.train_loader, self.test_loader, _, self.dataset_root = pypaddle.util.get_mnist_loaders(self.batch_size, self.possible_dataset_roots)
        self.device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        feature, labels = iter(self.train_loader).next()
        self.input_shape = feature.shape[1:]
        self.output_size = int(labels.shape[-1])

    def tearDown(self):
        if self.dataset_root is not self.possible_dataset_roots[0]:
            print('Deleting', self.dataset_root)
            shutil.rmtree(self.dataset_root)

    def test_develop(self):
        return

        # Arrange
        random_graph = nx.watts_strogatz_graph(30, 3, 0.8)
        structure = pypaddle.sparse.CachedLayeredGraph()
        structure.add_edges_from(random_graph.edges)
        structure.add_nodes_from(random_graph.nodes)

        # Depthwise Separable Convolutions
        class SeparableConv2d(nn.Module):
            def __init__(self, in_channels, out_channels, kernel_size=1, stride=1, padding=0, dilation=1, bias=False):
                super(SeparableConv2d, self).__init__()
                self.conv1 = nn.Conv2d(in_channels, in_channels, kernel_size, stride, padding, dilation, groups=in_channels, bias=bias)
                self.pointwise = nn.Conv2d(in_channels, out_channels, 1, 1, 0, 1, 1, bias=bias)

            def forward(self, x):
                x = self.conv1(x)
                x = self.pointwise(x)
                return x

        class CustomCell(nn.Module):
            def __init__(self, in_degree, in_channel, out_channel, stride):
                super(CustomCell, self).__init__()
                self.single = (in_degree == 1)
                if not self.single:
                    self.agg_weight = nn.Parameter(torch.zeros(in_degree, requires_grad=True))
                self.conv = SeparableConv2d(in_channel, out_channel, kernel_size=3, padding=1, stride=stride)
                self.bn = nn.BatchNorm2d(out_channel)
                self.act = nn.ReLU()
                self.num_out_channel = out_channel

            def forward(self, input):
                # input: [B, C_in, N, M, in_degree]
                if self.single:
                    input = input.squeeze(-1)
                else:
                    input = torch.matmul(input, torch.sigmoid(self.agg_weight))  # [B, C_in, N, M]
                input = self.conv(self.act(input))  # [B, C_out, N, M]
                input = self.bn(input)  # [B, C_out, N, M]
                return input

        class SkipMap(nn.Module):
            def __init__(self, in_channel, out_channel):
                super(SkipMap, self).__init__()
                #self.conv = nn.Conv2d(in_channel, out_channel, kernel_size=(3, 3), padding=(1, 1))
                self.conv = nn.Conv2d(in_channel, out_channel, kernel_size=1)
                self.bn = nn.BatchNorm2d(out_channel)
                self.act = nn.ReLU()

            def forward(self, input):
                # input: [B, C_in, N, M]
                input = self.conv(self.act(input))  # [B, C_out, N, M]
                input = self.bn(input)  # [B, C_out, N, M]
                return input

        def layer_channel_size(layer: int):
            # MNIST: input layer channel size 1
            # CIFAR: input layer channel size 3
            return 1 if layer is 0 else 40 if layer < 4 else 80 if layer < 8 else 160

        def construct_cell(is_input, is_output, in_degree, out_degree, layer):
            in_channel = layer_channel_size(0 if is_input else layer)
            in_degree = max(1, in_degree)  # shouldn't be zero
            out_channel = layer_channel_size(layer+1)
            cell = CustomCell(in_degree, in_channel, out_channel, 1)
            cell.layer = layer
            return cell

        model = paddle.deprecated.DeepDACellNetwork(10, construct_cell, layer_channel_size, SkipMap, structure)
        print(model)
        print([p.shape for p in model.parameters()])

        def count_parameters(model):
            return sum(p.numel() for p in model.parameters() if p.requires_grad)

        # download and transform train dataset
        mnist_train_loader = self.train_loader
        mnist_test_loader = self.test_loader

        device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        print(device)
        model.to(device)

        #optimizer = torch.optim.SGD(model.parameters(), lr=0.01, momentum=0.5)
        optimizer = torch.optim.Adam(model.parameters(), lr=0.01, betas=(0.9, 0.999), eps=1e-8, weight_decay=0)
        loss_func = nn.CrossEntropyLoss()

        for epoch in range(250):
            print('Epoch', epoch)
            model.train()
            print(train(mnist_train_loader, model, optimizer, loss_func, device))

            model.eval()
            print(test(mnist_test_loader, model, device))

        print('Done')
        print('Test:')
        model.eval()
        print(test(mnist_test_loader, model, device))


import os
# Setuptools should be preferred:
# * https://stackoverflow.com/a/25372045/5578392
# * https://stackoverflow.com/a/14753678/5578392
#
# https://stackoverflow.com/a/41573588/5578392
from setuptools import setup
from setuptools import find_packages

with open(os.path.join('./', 'VERSION')) as version_file:
    project_version = version_file.read().strip()

setup(
    name='pypaddle',
    version=project_version,
    description='Innvariant Group Sparse Neural Network Tools',
    author='Chair of Data Science, University of Passau',
    author_email='julian.stier@uni-passau.de',
    url='https://github.com/innvariant/pypaddle',
    download_url = 'https://github.com/innvariant/pypaddle/archive/%s.tar.gz' % project_version,
    keywords=['sparse', 'neural', 'networks', 'graph theory', 'pytorch'],
    packages=find_packages(),
    license="GPL3 License",
    install_requires=['pip', 'setuptools>=18.0', 'torch>=1.3', 'networkx'],
    dependency_links=[],
    classifiers = [
        "Development Status :: 1 - Planning",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: Other/Proprietary License",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
)


import numpy as np
import torch
import unittest
import pypaddle.sparse
import pypaddle.util


class MaskedLinearLayerTest(unittest.TestCase):
    def test_set_mask_explicitly_success(self):
        input_size = 5
        output_size = 2
        layer = pypaddle.sparse.MaskedLinearLayer(input_size, output_size)
        mask = torch.zeros((output_size, input_size), dtype=torch.bool)
        mask[0, 0] = 1
        mask[0, 1] = 1
        mask[1, 2] = 1

        layer.set_mask(mask)

        self.assertTrue(np.all(np.equal(np.array(mask), np.array(layer.mask))))

    def test_parameter_reset_success(self):
        # Arrange - initialize a masked layer and randomize its mask
        input_size = 5
        output_size = 7
        layer = pypaddle.sparse.MaskedLinearLayer(input_size, output_size)
        layer.apply(pypaddle.util.set_random_masks)
        initial_state = np.copy(layer.mask)

        # Act - Now the mask should be reset to only ones
        layer.reset_parameters()

        # Assert - The random mask and the resetted mask should not match
        self.assertEqual(layer.mask.size(), initial_state.shape)
        self.assertTrue((np.array(layer.mask) != initial_state).any())

    def test_mask_changes_output_success(self):
        input_size = 5
        output_size = 7
        layer = pypaddle.sparse.MaskedLinearLayer(input_size, output_size)
        input = torch.rand(input_size)

        layer.apply(pypaddle.util.set_random_masks)
        first_mask = np.copy(layer.mask)
        first_mask_output = layer(input).detach().numpy()
        layer.apply(pypaddle.util.set_random_masks)
        second_mask = np.copy(layer.mask)
        second_mask_output = layer(input).detach().numpy()

        self.assertTrue((first_mask != second_mask).any(), 'Masks for inference should not equal, but are randomly generated.')
        self.assertTrue(np.any(np.not_equal(first_mask_output, second_mask_output)))

    def test_random_input_success(self):
        input_size = 5
        output_size = 2
        model = pypaddle.sparse.MaskedLinearLayer(input_size, output_size)
        input = torch.tensor(np.random.random(input_size))

        output = model(input)

        self.assertEqual(output.numel(), output_size)